# This code briefly tested and appears to be working correctly, deserves more testing though; Hader 2018/08/06.
#
# This script manipulates a suite of CAL3QHC input files by
# generating new input files for combinations of user-specified
# vehicle emission rates and stability classes/surface roughnesses.
# The lines of the CAL3QHC input file (*.in2) files are modified as follows:
#
# First line: The Job title is modified to reflect the input parameters,
# and the surface roughness is modified to reflect the urban/rural classification.
# Line "3": Line after the receptors. The Run title is set to be the same as the job title.
# Link lines: start Line number defined by the number of receptors plus 4 for offset,
# then every other line for the number of links. The emission values are modified in 
# these lines, dependent on whether the link is up or downhill and rural or urban.
# Line '6': i.e., the line immediately following the last link definition.
# The stability class is modified to reflect the urban/rural classicication.
#
# Additional notes:
# The input file path must contain only *.in2 input files that will have the
# combinations of various parameters built into new files. These base file names
# must be of the form "CAL3QHC_##lanes.in2"
#
# Author: John Hader

# Clears the memory and the text on the console.
rm(list = ls())
cat("\014")

# Loads required package(s).
require(qdapRegex)

# User-defined inputs -----------------------------------------------------

# Input/output paths and base file definition.
input_path <- "C:/Users/37880/Desktop/NCHRP/Runs/20%_Trucks/Freeways/02_Base_input_files" # Contains only .in2 files. 
output_path <- "C:/Users/37880/Desktop/NCHRP/Runs/20%_Trucks/Freeways/03_Full_input_files/CAL3QHC_" # Full file names defined in script.

# Input file specifications.
num_links <- 2
link_grade_direction <- c("-","+") # "-" defines downhill, "+" defines uphill. Emission rates will be selected depending on these definitions.
num_receptors <- 20
surf_rough_location <- 2 # Parameter number on line 1, excluding the Job Title.
emiss_rate_location <- 7 # Parameter number in the link lines, excluding link name, where the emission rate is defined.
stab_class_location <- 4 # Parameter number in last line where stability class is, plus 1 (first param  = "").

# Emission rate values. One for positive and one for 
# negative grades. Input as matrices w/first column denoting urban and 
# second denoting rural emissions for corresponding grades 
# (grades defined down the rows of the matrix).
emiss_rates_positive <- matrix(data=c(6.17,8.63,11.35,13.84,16.56,19.39,21.60,23.07,
                                      6.17,8.62,11.34,13.81,16.52,19.34,21.54,23.00),nrow=8,ncol=2,byrow=FALSE)
emiss_rates_negative <- matrix(data=c(6.17,4.52,3.53,2.84,2.29,1.69,1.10,0.77,
                                      6.17,4.53,3.54,2.85,2.31,1.70,1.10,0.77),nrow=8,ncol=2,byrow=FALSE)
emiss_rates_labels <- c("0%","1%","2%","3%","4%","5%","6%","7%") # "+" and "-" added in code.

# Urban vs. Rural parameters and names. Note that these MUST be added as c(Urban,Rural),
# so-as to match the format of the emissions values (i.e., first column is urban).
stab_classes <- c("4","5") # Class 4= D, Class 5= E.
surf_rough <- c(108,1) # Must be integers
urban_rural_labels <- c("U","R")

#------ End User-defined Inputs--------

# Pre-processing ----------------------------------------------------------

# Checks that emiss rates defined correctly.
if((nrow(emiss_rates_positive)!=length(emiss_rates_labels))|(nrow(emiss_rates_negative)!=length(emiss_rates_labels))){
  stop("The emission rates matrices must each have the same number of rows as there are emiss_rates_labels.")
}

# Quasi-static input parameters.
num_lines_to_link_01 <- 4+num_receptors
string_param <- "'" # Character bounding sting parameters in the .in2 file.
num_lines_to_run_title <- 2+num_receptors
num_lines_per_link <- 2

# Reads in base files and names.
input_file_paths <- list.files(path=input_path,full.names=TRUE)
input_file_names <- list.files(path=input_path,full.names=FALSE)

# Extracts number of lanes from input files.
input_file_lanes <- as.numeric(unlist(rm_between(input_file_names,left="_",right="_lanes",fixed=TRUE,extract=TRUE)))

# Files processing --------------------------------------------------------

# Loops through the input files.
for(i in c(1:length(input_file_paths))){
  
  # Reads in the current file.
  current_file <- readLines(input_file_paths[i],warn=TRUE,skipNul=FALSE)
  current_lanes <- input_file_lanes[i]
  
  # Loops through the emission rates (i.e., grades).
  for(j in c(1:nrow(emiss_rates_positive))){
    
    # Defines the current emission rate labels (values extracted in U/R loop).
    current_emiss_label <- emiss_rates_labels[j]
    
    # Loops through the urban/rural parameters.
    for(k in c(1:length(urban_rural_labels))){
      
      # Generates copy of the current file data.
      current_file_mod <- current_file
      
       # Defines the current urban/rural parameters to implement.
      current_stab_class <- stab_classes[k]
      current_surf_rough <- surf_rough[k]
      current_urb_rur_label <- urban_rural_labels[k]
      
      # Defines job title and ensures length is acceptable.
      current_job_title <- paste(string_param,"CAL3QHC ",current_lanes," lanes ",
                                 current_emiss_label," grade ",current_urb_rur_label,string_param,sep="")
      if(nchar(current_job_title)>40){
        stop("The job title for the current input file is too long. Modify the parameter labels so that the length is no more than 40 characters.")
      }

      # Changes the label and surface roughness of the file (line 1).# --------

      # Extracts the parts of line 1 that need modification.
      line_01 <- current_file_mod[1]
      line_01_params_all <- substr(line_01,start=2+gregexpr(pattern=string_param,line_01)[[1]][2],stop=nchar(line_01))
      #
      # Extracts the numerical parameters separately.
      line_01_params_separate <- unlist(strsplit(line_01_params_all, "\\s+"))
      #
      # Replaces job title and surf rough w/current values, and replaces line in data.
      line_01_params_separate[surf_rough_location] <- paste(current_surf_rough,".000",sep="")
      line_01_revised <- paste(current_job_title,paste(line_01_params_separate, collapse=" "))
      current_file_mod[1] <- line_01_revised
      
      # Changes the run title of the file to the same as the job title (line "3" (after all receptors)).# -------- 
      run_title_line <- current_file_mod[num_lines_to_run_title]
      run_title_line_params <- substr(run_title_line,start=2+gregexpr(pattern=string_param,run_title_line)[[1]][2],stop=nchar(run_title_line))
      run_title_line <- paste(current_job_title,run_title_line_params,sep=" ")
      current_file_mod[num_lines_to_run_title] <- run_title_line
        
      # Loops through links, setting emiss. values as needed -------------------- 
      
      # Loops through links.
      for(l in c(1:num_links)){
        
        # Extracts the current emission rate for the current U/R setting,
        # depending on whether the link is uphill or downhill.
        if(link_grade_direction[l]=="+"){
          emiss_rates_to_use <- emiss_rates_positive
        } else if(link_grade_direction[l]=="-"){
          emiss_rates_to_use <- emiss_rates_negative
        } else{
          stop("An incorrect link_grade_direction parameter was defined by the user. These must be wither + or -.")
        }
        current_emiss_rate <- emiss_rates_to_use[j,k]
        
        # Extracts the current link name and parameters.
        current_link_line <- current_file_mod[num_lines_to_link_01+((l-1)*num_lines_per_link)]
        current_link_name <- substr(current_link_line,start=1,stop=gregexpr(pattern=string_param,current_link_line)[[1]][2])
        current_link_params <- substr(current_link_line,start=2+gregexpr(pattern=string_param,current_link_line)[[1]][2],stop=nchar(current_link_line))
        
        # Extracts the link parameters separately.
        current_link_params_separate <- unlist(strsplit(current_link_params,"\\s+"))
        #
        # Inserts current emission rate, and replaces line.
        current_link_params_separate[emiss_rate_location] <- current_emiss_rate
        current_link_line_revised <- paste(current_link_name,paste(current_link_params_separate, collapse=" "))
        current_file_mod[num_lines_to_link_01+((l-1)*num_lines_per_link)] <- current_link_line_revised

      }
      
      # Final line setting and file printing --------------------       
      
      # Modifies the last line of the file (contains meteorology parameter).
      last_line <- current_file_mod[length(current_file_mod)]
      last_line_params_separate <- unlist(strsplit(last_line,"\\s+"))
      last_line_params_separate[stab_class_location] <- current_stab_class
      
      # Inserts the corrected last line of the file.
      last_line_revised <- paste(last_line_params_separate, collapse=" ")
      current_file_mod[length(current_file_mod)] <- last_line_revised

      # Construct file name from current parameters.
      out_file_name <- paste(output_path,sprintf("%02d",current_lanes),"lanes_",current_emiss_label,
                             "_",current_urb_rur_label,".in2",sep="")
      
      # Writes out the modified text data.
      write(current_file_mod, file=out_file_name)

    }
  }
}

# This script was briefly tested by inputting a paired-down version of the input data, and manually spot-checking
# the output files using simplified input parameters. No issues were detected with these spot-checks, but
# further testing would be warrented.
#
# Hader; 2018/08/06.
