# Minimally spot-checked by Hader; 2018/09/20
#
# This script manipulates a suite of CAL3QHC input files by
# generating new input files for combinations of user-specified
# vehicle emission rates, stability classes/surface roughnesses, 
# and a suite of highway-to-intersection set-back distances.
# The lines of the CAL3QHC input file (*.in2) files are modified as follows:
#
# First line: The Job title is modified to reflect the input parameters,
# and the surface roughness is modified to reflect the urban/rural classification.
#
# Line "2": Receptor Lines: All receptors with a name that begins with a user-definition
# (typically, 'HW Right' will have their x-coordinates shifted by the appropriate amount
# to account for the specific highway setback distance from the intersection).
# This effectively sets the highway the given distance back from the 'ghost'
# intersection receptors (i.e., those reflective of if the highway were on the other side
# of the intersection).
# 
# Line "3": Line after the receptors. The Run title is set to be the same as the job title.
#
# Link lines: start Line number defined by the number of receptors plus 4 for offset
# (to skip the '1' delineation). Then, the first line of the link is manipulated as follows:
#
# The emissions factor (g/veh-mi) is modified to reflect
# the given urban/rural setting (this script does not handle emission rates for varying
# gradients/directions up or downhill).
# Also, the X-coordinates of the link are changed 
# to reflect the new setback distance from the intersection (this effectively sets the 
# highway the given distance back from the 'real' intersection receptors).
#
# Line '6': i.e., the line immediately following the last link definition.
# The stability class is modified to reflect the urban/rural classification.
#
# Additional notes:
# The input file path directory must contain only *.in2 input files that will have the
# combinations of various parameters built into new files. These base file names
# must be of the form "CAL3QHC_##_deg###_##_###.in2", where ## refers to the skew angle
# of the coordinates in the input file.
# Also, no change in slope gradient is taken into account in this processing.
#
# Author: John Hader

# Clears the memory and the text on the console.
rm(list = ls())
cat("\014")

# Loads required package(s).
require(qdapRegex)

# User-defined inputs -----------------------------------------------------

# Input/output paths.
#
# QA'ed (JDH; 2019/11/01)
input_path <- "C:/Users/37880/Desktop/NCHRP/Runs/20%_Trucks/Interchanges_2019/02_Base_input_files_FW" # Contains only .in2 files. 
output_path <- "C:/Users/37880/Desktop/NCHRP/Runs/20%_Trucks/Interchanges_2019/03_Full_input_files_FW/CAL3QHC_" # Full file names defined in script.

# Establishes locations of parameters in input file names (these positions are after strpslt w/ "_").
#
# QA'ed (JDH; 2019/11/01)
skew_loc <- 2
lane_loc <- 4
dist_loc <- 6

# Input file specifications.
#
# QA'ed (JDH; 2019/11/01)
num_links <- 2
num_receptors <- 8
surf_rough_location <- 2 # Parameter number on line 1, excluding the Job Title.
rec_x_coord_location <- 1 # Parameter number in receptor line for x-coord, excluding rec. name.
link_x_coords_locations <- c(2,4) # Parameter number in link lines where x-coordinates are, excluding link name.
emiss_rate_location <- 7 # Parameter number in the link lines, excluding link name, where the emission rate is defined.
stab_class_location <- 3 # Parameter number in last line where stability class is.

# Emission rate values. First element is for
# urban scenarios, second is for rural.
#
# QA'ed (JDH; 2019/11/01)
emiss_rates <- c(6.17,6.17) # In grams/veh-mile
gradient <- "0%"

# Defines several parameters associated with varying the highway setback distance.
#
# QA'ed (JDH; 2019/11/01)
set_back_distances_labels <- c(20,30,60,80,100,125,150,175,300,500,1000) # Units in feet. Note these will be adjusted by the initial distance of 20ft!
initial_distance <- 20 # Units in feet.
ft_to_m_factor <- 0.3048
receptors_to_setback <- "'HW Right" # Rec names that start w/this (include the ') will have the x-coord adjusted (negatively) by the setback dist.

# Urban vs. Rural parameters and names. Note that these MUST be added as c(Urban,Rural),
# so-as to match the format of the emissions values (i.e., first column is urban).
#
# QA'ed (JDH; 2019/11/01)
stab_classes <- c("4","5") # Class 4= D, Class 5= E.
surf_rough <- c(108,1) # Must be integers
urban_rural_labels <- c("U","R")

#------ End User-defined Inputs--------

# Pre-processing ----------------------------------------------------------

# Checks that emiss rates defined correctly.
if(length(emiss_rates)!=length(urban_rural_labels)){
  stop("The number of emission rates must match the number of urban/rural labels.")
}

# Quasi-static input parameters.
num_lines_to_link_01 <- 4+num_receptors
string_param <- "'" # Character bounding sting parameters in the .in2 file.
num_lines_to_run_title <- 2+num_receptors
num_lines_per_link <- 2

# Reads in base files and names.
input_file_paths <- list.files(path=input_path,full.names=TRUE)
input_file_names <- list.files(path=input_path,full.names=FALSE)

# Converts setback distances to meters relative to starting setback distance
# this is because the receptor coords are both in meters, and already have the starting
# setback distance reflected in them.
set_back_distances <- set_back_distances_labels-initial_distance
set_back_distances <- set_back_distances*ft_to_m_factor

# Files processing --------------------------------------------------------

# Loops through the input files.
for(i in c(1:length(input_file_paths))){
  
  # Reads in the current file.
  current_file <- readLines(input_file_paths[i],warn=TRUE,skipNul=FALSE)
  current_file_name <- input_file_names[i]
  
  # Establishes parameters from the current file.
  file_name_params <- strsplit(current_file_name,"_")[[1]]
  current_deg <- file_name_params[skew_loc]
  current_lanes <- file_name_params[lane_loc]
  
  # Ensures setback distance matches user-definition.
  if(as.numeric(file_name_params[dist_loc])!=initial_distance){
    stop(paste("User defined initial setback distance of",intial_distance,"does not equal that in file #",i,"."))
  }
  
  # Loops through the urban/rural parameters.
  for(k in c(1:length(urban_rural_labels))){
    
    # Defnes urban vs. rural parameters to implement.
    current_emiss_rate <- emiss_rates[k]
    current_stab_class <- stab_classes[k]
    current_surf_rough <- surf_rough[k]
    current_urb_rur_label <- urban_rural_labels[k]
    
    # Loops through the set-back distances of the highway from the intersection.
    for(l in c(1:length(set_back_distances))){
      
      # Extracts current set back distance.
      current_setback <- set_back_distances[l]
      
      # Generates copy of the current file data.
      current_file_mod <- current_file
      
      # Defines job title and ensures length is acceptable.
      current_job_title <- paste(string_param,"FW ",current_deg," deg ",gradient," grade ",current_urb_rur_label," ",
                                 current_lanes," ln ",set_back_distances_labels[l]," ft",string_param,sep="")
      if(nchar(current_job_title)>40){
        stop("The job title for the current input file is too long. Modify the parameter labels so that the length is no more than 40 characters.")
      }
      
      # Changes Line 1 parameters --------
      
      # Extracts the parts of line 1 that need modification.
      line_01 <- current_file_mod[1]
      line_01_params_all <- substr(line_01,start=2+gregexpr(pattern=string_param,line_01)[[1]][2],stop=nchar(line_01))
      #
      # Extracts the numerical parameters separately.
      line_01_params_separate <- unlist(strsplit(line_01_params_all, ","))
      #
      # Replaces job title and surf rough w/current values, and replaces line in data.
      line_01_params_separate[surf_rough_location] <- paste(current_surf_rough,".000",sep="")
      line_01_revised <- paste(current_job_title,paste(line_01_params_separate,collapse=","),sep=",")
      current_file_mod[1] <- line_01_revised
      
      # Loops through the receptors, and modifies the x-coord of receptors as necessary.
      # Note, m index is shifted by one to account for receptors starting on line 2.
      for(m in c(2:(num_receptors+1))){
        
        # Extracts current receptor line from working file.
        current_rec_line <- current_file_mod[m]
        
        # Determines if rec. needs x-coord modified.
        rec_name <- substr(current_rec_line,start=1,stop=gregexpr(pattern=string_param,current_rec_line)[[1]][2])
        if(identical(substr(rec_name,start=1,stop=nchar(receptors_to_setback)),receptors_to_setback)){
          # Extracts rec. params.
          rec_params <- substr(current_rec_line,start=2+gregexpr(pattern=string_param,current_rec_line)[[1]][2],stop=nchar(current_rec_line))
          rec_params_sep <- unlist(strsplit(rec_params,","))
          
          # Adjusts x-coord by necessary setback (multiplied by 2 to account for the links being shifted as well as the necessary increase setback from these links).
          rec_params_sep[rec_x_coord_location] <- sprintf("%.5f",(as.numeric(rec_params_sep[rec_x_coord_location])-(2*current_setback)))
          current_file_mod[m] <- paste(rec_name,paste(rec_params_sep,collapse=","),sep=",")
          
        } else{
          # Nothing is done if the rec. name doesn't match the user definition.
        }
        
      }
      
      # Changes the run title of the file to the same as the job title (line "3" (after all receptors)).# -------- 
      run_title_line <- current_file_mod[num_lines_to_run_title]
      run_title_line_params <- substr(run_title_line,start=2+gregexpr(pattern=string_param,run_title_line)[[1]][2],stop=nchar(run_title_line))
      run_title_line <- paste(current_job_title,run_title_line_params,sep=",")
      current_file_mod[num_lines_to_run_title] <- run_title_line
      
      # Loops through links, setting x coord and emiss. values as needed -------------------- 
      
      # Establishes the line counter for looping through the links.
      current_line_num <- num_lines_to_link_01
      
      # Loops through links.
      for(m in c(1:num_links)){
        
        # Extracts the current link name and parameters.
        current_link_line <- current_file_mod[current_line_num]
        current_link_name <- substr(current_link_line,start=1,stop=gregexpr(pattern=string_param,current_link_line)[[1]][2])
        current_link_params <- substr(current_link_line,start=2+gregexpr(pattern=string_param,current_link_line)[[1]][2],stop=nchar(current_link_line))
        
        # Extracts the link parameters separately.
        current_link_params_sep <- unlist(strsplit(current_link_params,","))
        #
        # Inserts current emission rate.
        current_link_params_sep[emiss_rate_location] <- current_emiss_rate
        
        # Modifies x-coord of link by setback distance.
        # Link start coord.
        #
        current_link_params_sep[link_x_coords_locations[1]] <- 
          sprintf("%.5f",(as.numeric(current_link_params_sep[link_x_coords_locations[1]])-current_setback))
        # Link stop coord
        #
        current_link_params_sep[link_x_coords_locations[2]] <- 
          sprintf("%.5f",(as.numeric(current_link_params_sep[link_x_coords_locations[2]])-current_setback))
        
        # Replaces link line in working file.
        current_link_line_revised <- paste(current_link_name,paste(current_link_params_sep,collapse=","),sep=",")
        current_file_mod[current_line_num] <- current_link_line_revised
        
        # Increments current line number for next link.
        current_line_num <- current_line_num+num_lines_per_link
      }
      
      # Final line setting and file printing --------------------       
      
      # Modifies the last line of the file (contains meteorology parameter).
      last_line <- current_file_mod[length(current_file_mod)]
      last_line_params_separate <- unlist(strsplit(last_line,","))
      last_line_params_separate[stab_class_location] <- current_stab_class
      
      # Inserts the corrected last line of the file.
      last_line_revised <- paste(last_line_params_separate,collapse=",")
      current_file_mod[length(current_file_mod)] <- last_line_revised
      
      # Construct file name from current parameters.
      out_file_name <- paste(output_path,current_deg,"_deg",gradient,"_",current_lanes,"ln_",
                             set_back_distances_labels[l],"ft_",current_urb_rur_label,".in2",sep="")
      
      # Writes out the modified text data.
      write(current_file_mod,file=out_file_name)
      
    }
    
  }
}

# This script was reviewed by Hader after writing, along w/spot-checking of results in the first-pass file.
# After running for the 10% and 20% trucks files, the output from files was spot-check compared to
# coordinate values produced by the Excel builder (which generates the coordinates for the links
# and HW Right receptors). Various files were also inter-compared to ensure that certain parameters
# were identical between files where expected given scenario specifics (i.e., urban vs. rural, 
# 10% vs. 20%, corresponding files for different skew angles).
#
# Hader; 2018/09/20