# 
#
# This script manipulates a suite of CAL3QHC input files by
# generating new input files for combinations of user-specified
# vehicle emission rates and stability classes/surface roughnesses.
# The lines of the CAL3QHC input file (*.in2) files are modified as follows:
#
# First line: The Job title is modified to reflect the input parameters,
# and the surface roughness is modified to reflect the urban/rural classification.
#
# Line "3": Line after the receptors. The Run title is set to be the same as the job title.
#
# Link lines: start Line number defined by the number of receptors plus 4 for offset
# (to skip the 1/2 delineation). Then, based on user definitions, either the first (for
# free-flow links) or second (for queue links) line of the link is manipulated as follows:
#
# Free-flow (1 line): The emissions factor (g/veh-mi) are modified to reflect
# the given grade, rural setting, and whether the link is uphill or downhill.
#
# Queue (2 lines): First line has nothing done to it. Second line has the
# idle emission factor modified to reflect the urban/rural setting.
#
# Line '6': i.e., the line immediately following the last link definition.
# The stability class is modified to reflect the urban/rural classification.
#
# Additional notes:
# The input file path must contain only *.in2 input files that will have the
# combinations of various parameters built into new files. These base file names
# must be of the form "CAL3QHC_##_deg.in2", where ## refers to the skew angle
# of the coordinates in the input file.
#
# Author: John Hader

# Clears the memory and the text on the console.
rm(list = ls())
cat("\014")

# Loads required package(s).
require(qdapRegex)

# User-defined inputs -----------------------------------------------------

# Input/output paths.
#
# QA'ed all parameters (JDH, 2019/01/11)
input_path <- "C:/Users/37880/Desktop/NCHRP/Runs/10%_Trucks/Interchanges_2019/02_Base_input_files_IN" # Contains only .in2 files. 
output_path <- "C:/Users/37880/Desktop/NCHRP/Runs/10%_Trucks/Interchanges_2019/03_Full_input_files_IN/CAL3QHC_" # Full file names defined in script.

# Input file specifications.
# N and W bound are uphill
#
# QA'ed all parameters (JDH, 2019/01/11)
num_links <- 16
link_grade_direction <- c("+","+","+","+","-","-","-","-","-","-","-","-","+","+","+","+") # Order corresponds to link order in input file.
num_receptors <- 4
surf_rough_location <- 2 # Parameter number on line 1, excluding the Job Title.
FF_emiss_rate_location <- 7 # Parameter number in the free-flow link lines, excluding link name, where the emission rate is defined.
queue_emiss_param_loc <- 5 # Parameter number in second queue line that defines emission rate.
stab_class_location <- 3 # Parameter number in last line where stability class is.

# Moving emission rate values. One for positive and one for 
# negative grades. Input as matrices w/first column denoting urban and 
# second denoting rural emissions for corresponding grades 
# (grades defined down the rows of the matrix).
# For intersections, speed used is 15 mph; urban and rural unrestricted. 
# Emissions are in grams/vehi-mile. 
#
# QA'ed all parameters (JDH, 2019/01/11)
emiss_rates_positive <- matrix(data=c(6.60, 7.31,8.18,8.95,10.09,11.25,12.94,14.49,
                                      6.62,7.34,8.20,8.97,10.11,11.27,12.96,14.51),nrow=8,ncol=2,byrow=FALSE)
emiss_rates_negative <- matrix(data=c(6.60,6.00,5.49,5.04,4.70,4.24,3.72,3.45,
                                      6.62,6.02,5.51,5.05,4.71,4.26,3.73,3.47),nrow=8,ncol=2,byrow=FALSE)
emiss_rates_labels <- c("0%","1%","2%","3%","4%","5%","6%","7%") # "+" and "-" added in code.

# Idling emission rates (for queue links). First element = Urban, second = Rural.
# Units are in grams/veh-hour.
#
# QA'ed (Hader; 2019/01/11)
idling_emiss_rates <- c(11.42,11.58)

# Vector defining type of links (in order). Needed because queue links (denote as 2)
# have 2 lines, the second of which is modified, while freeflow links (denote as 1)
# have 1 line and are modified.
#
# QA'ed (Hader; 2019/09/01)
free_vs_queue <- c(1,2,2,1,1,2,2,1,1,2,2,1,1,2,2,1)

# Urban vs. Rural parameters and names. Note that these MUST be added as c(Urban,Rural),
# so-as to match the format of the emissions values (i.e., first column is urban).
#
# QA'ed (Hader; 2019/09/01)
stab_classes <- c("4","5") # Class 4= D, Class 5= E.
surf_rough <- c(108,1) # Must be integers; units are in cm.
urban_rural_labels <- c("U","R")

#------ End User-defined Inputs--------

# Pre-processing ----------------------------------------------------------

# Checks that emiss rates defined correctly.
if((nrow(emiss_rates_positive)!=length(emiss_rates_labels))|(nrow(emiss_rates_negative)!=length(emiss_rates_labels))){
  stop("The emission rates matrices must each have the same number of rows as there are emiss_rates_labels.")
}

# Quasi-static input parameters.
num_lines_to_link_01 <- 4+num_receptors
string_param <- "'" # Character bounding sting parameters in the .in2 file.
num_lines_to_run_title <- 2+num_receptors

# Reads in base files and names.
input_file_paths <- list.files(path=input_path,full.names=TRUE)
input_file_names <- list.files(path=input_path,full.names=FALSE)

# Extracts skew angle degree from input files.
input_file_skews <- as.numeric(unlist(rm_between(input_file_names,left="_",right="_deg",fixed=TRUE,extract=TRUE)))

# Files processing --------------------------------------------------------

# Loops through the input files.
for(i in c(1:length(input_file_paths))){
  
  # Reads in the current file.
  current_file <- readLines(input_file_paths[i],warn=TRUE,skipNul=FALSE)
  current_skew <- input_file_skews[i]
  
  # Loops through the emission rates (i.e., grades).
  for(j in c(1:nrow(emiss_rates_positive))){
    
    # Defines the current emission rate labels (values extracted in U/R loop).
    current_emiss_label <- emiss_rates_labels[j]
    
    # Loops through the urban/rural parameters.
    for(k in c(1:length(urban_rural_labels))){
      
      # Generates copy of the current file data.
      current_file_mod <- current_file
      
       # Defines the current urban/rural parameters to implement.
      current_stab_class <- stab_classes[k]
      current_surf_rough <- surf_rough[k]
      current_urb_rur_label <- urban_rural_labels[k]
      
      # Defines job title and ensures length is acceptable.
      current_job_title <- paste(string_param,"CAL3QHC ",current_skew," deg ",
                                 current_emiss_label," grade ",current_urb_rur_label,string_param,sep="")
      if(nchar(current_job_title)>40){
        stop("The job title for the current input file is too long. Modify the parameter labels so that the length is no more than 40 characters.")
      }

      # Changes the label and surface roughness of the file (line 1).# --------

      # Extracts the parts of line 1 that need modification.
      line_01 <- current_file_mod[1]
      line_01_params_all <- substr(line_01,start=2+gregexpr(pattern=string_param,line_01)[[1]][2],stop=nchar(line_01))
      #
      # Extracts the numerical parameters separately.
      line_01_params_separate <- unlist(strsplit(line_01_params_all, ","))
      #
      # Replaces job title and surf rough w/current values, and replaces line in data.
      line_01_params_separate[surf_rough_location] <- paste(current_surf_rough,".000",sep="")
      line_01_revised <- paste(current_job_title,paste(line_01_params_separate, collapse=","),sep=",")
      current_file_mod[1] <- line_01_revised
      
      # Changes the run title of the file to the same as the job title (line "3" (after all receptors)).# -------- 
      run_title_line <- current_file_mod[num_lines_to_run_title]
      run_title_line_params <- substr(run_title_line,start=2+gregexpr(pattern=string_param,run_title_line)[[1]][2],stop=nchar(run_title_line))
      run_title_line <- paste(current_job_title,run_title_line_params,sep=",")
      current_file_mod[num_lines_to_run_title] <- run_title_line
        
      # Loops through links, setting emiss. values as needed -------------------- 
      
      # Establishes the line counter for looping through the links.
      current_line_num <- num_lines_to_link_01
      
      # Loops through links.
      for(l in c(1:num_links)){
        
        # Extracts information for the current link
        current_link_grade_dir <- link_grade_direction[l]
        current_link_type <- free_vs_queue[l]
      
        # Extracts emission rate for current link type and rurality.
        if(current_link_type==1){
         
          # Extracts emiss rate for uphill vs downhill and U/R.
          if(current_link_grade_dir=="+"){
            emiss_rates_to_use <- emiss_rates_positive
          } else if(current_link_grade_dir=="-"){
            emiss_rates_to_use <- emiss_rates_negative
          } else{
            stop("An incorrect link_grade_direction parameter was defined by the user. These must be wither + or -.")
          }
          current_emiss_rate <- emiss_rates_to_use[j,k]
          
        } else if(current_link_type==2){
          
          # Exctracts idling rate for urban vs. rural
          current_emiss_rate <- idling_emiss_rates[k]
          
          # Modifies the line counter as necessary (i.e., to skip to the second line of the link).
          current_line_num <- current_line_num+1
          
        } else{
          stop("An incorrect free_vs_queue parameter was defined: Only 1 or 2 is allowed.")
        }
          
        # Handles link line modifications differently for queue vs. freeflow.
        if(current_link_type==1){
          
          # Extracts the current link name and parameters.
          current_link_line <- current_file_mod[current_line_num]
          current_link_name <- substr(current_link_line,start=1,stop=gregexpr(pattern=string_param,current_link_line)[[1]][2])
          current_link_params <- substr(current_link_line,start=2+gregexpr(pattern=string_param,current_link_line)[[1]][2],stop=nchar(current_link_line))
          
          # Extracts the link parameters separately.
          current_link_params_separate <- unlist(strsplit(current_link_params,","))
          #
          # Inserts current emission rate, and replaces line.
          current_link_params_separate[FF_emiss_rate_location] <- current_emiss_rate
          current_link_line_revised <- paste(current_link_name,paste(current_link_params_separate,collapse=","),sep=",")
          current_file_mod[current_line_num] <- current_link_line_revised
          
        } else if(current_link_type==2){
          
          # Extracts info on the second line of the queue link
          current_link_line <- current_file_mod[current_line_num]
          current_link_params_separate <- unlist(strsplit(current_link_line,","))
          
          # Inserts current idling emisson rate, and replaces line in revising file.
          current_link_params_separate[queue_emiss_param_loc] <- current_emiss_rate
          
          # Replaces link line 2 in the file.
          current_link_line_revised <- paste(current_link_params_separate, collapse=",")
          current_file_mod[current_line_num] <- current_link_line_revised
          
        }
        
        # Increments the line counter as necessary.
        current_line_num <- current_line_num+2
      }
      
      # Final line setting and file printing --------------------       
      
      # Modifies the last line of the file (contains a U/R meteorology parameter).
      last_line <- current_file_mod[length(current_file_mod)]
      last_line_params_separate <- unlist(strsplit(last_line,","))
      last_line_params_separate[stab_class_location] <- current_stab_class
      
      # Inserts the corrected last line of the file.
      last_line_revised <- paste(last_line_params_separate, collapse=",")
      current_file_mod[length(current_file_mod)] <- last_line_revised

      # Construct file name from current parameters.
      out_file_name <- paste(output_path,sprintf("%02d",current_skew),"_deg",current_emiss_label,
                             "_",current_urb_rur_label,".in2",sep="")
      
      # Writes out the modified text data.
      write(current_file_mod, file=out_file_name)

    }
  }
}

# This script was reviewed visually by Hader on 2019/01/09 after an NCHRP hiatus. Additionally, this script was 
# run on the data files for the intersections modeling, and ....
#
# Hader; 