# 
#
# This script reads in a series of .ou2 files from 
# the CAL3QHC model, extracts the line that contains
# info about the highest concentration from any receptor, 
# and outputs this data, along with the run info, to a .csv file.
# The .ou2 files must correspond to separate numbers
# of lanes (or skew angles), grades, and urban/rural classifications.
#
# Author: John Hader

# Clears memory, console, and loads package(s).
rm(list = ls())
cat("\014")
require(data.table)

# User-defined inputs -----------------------------------------------------

# Input folder and output file path.
#
# QA'ed (Hader; 2019/10/01)
base_dir <- "C:/Users/37880/Desktop/NCHRP/Runs/20%_Trucks/Intersections_2019/04_Output_files" # Move the .bat file used in CAL3QHC runs to different folder before running this script!
output_path <- "C:/Users/37880/Desktop/NCHRP/Runs/20%_Trucks/Intersections_2019/Intersections_20%_Results.csv" # As a .csv file.
file_name_sep <- "_" # Character that separates pieces of input file names.
expected_units <- "PPM" # All caps.

# Line of interest info.
#
# QA'ed (Hader; 2019/10/01)
pre_conc_text <- " THE HIGHEST CONCENTRATION OF" # Case-sensitive. This text used to search for line of interest.
conc_pos <- 6 # These values used after line is strplit, assumes spaces are between items in the line. Add 1 to each to account for extra space at start.
units_pos <- 7
rec_pos <- 11

# Position of params in file names (after strsplit using file_name_sep).
#
# QA'ed (Hader; 2019/10/01)
is_intersection <- TRUE
lane_skew_pos <- 2 # First two chars of this parameter are assumed to contain the number of lanes or skew angle (i.e., the # lanes or skew angle must be in 2-digit format).
grade_pos <- 3 # "%" assumed to be present. If intersection, assumed that "deg" preceeds the grade angle and that grade angle is only 1-digit long.
rurality_pos <- 4 # Assumed to be last param in file name and one char long.

#------ End User-defined Inputs--------

# Files pre-processing ----------------------------------------------------

# Reads in files from directory.
all_files_paths <- list.files(path=base_dir,full.names=TRUE)
all_files_names <- list.files(path=base_dir,full.names=FALSE)

# Files processing --------------------------------------------------------

# Establishes column name of output file for lanes or skew angle.
if(is_intersection==TRUE){
  col_header <- "Skew_angle"
} else if(is_intersection==FALSE){
  col_header <- "Num_Lanes"
} else{
  stop("The is_intersection parameter must be either TRUE or FALSE.")
}

# Creates dataframe for storing  conc. info.
out_data <- as.data.frame(matrix(data=NA,nrow=length(all_files_paths),ncol=5))
colnames(out_data) <- c(col_header,"Grade_%","Urban_or_Rural",paste("Max_conc_",expected_units,sep=""),"Rec_Num")

# Loops through input files.
for(i in c(1:length(all_files_paths))){
  # Extracts run parameter info from current file name.
  file_params <- strsplit(all_files_names[i],file_name_sep)[[1]]
  current_lanes_or_skew <- as.numeric(substr(file_params[lane_skew_pos],start=1,stop=2))
  current_rurality <- substr(file_params[rurality_pos],start=1,stop=1) # Since file extension follows, takes first character of the last file name element.
  
  # Extracts the gradient, method depends on layout type.
  if(is_intersection==TRUE){
    temp_grade <- file_params[grade_pos]
    current_grade <- substring(temp_grade,4)
  } else if(is_intersection==FALSE){
    current_grade <- file_params[grade_pos]
  } else{
    stop("The is_intersection parameter must be either TRUE or FALSE.")
  }
  
  # Loads in current file and finds line of interest.
  current_file <- readLines(all_files_paths[i],warn=TRUE,skipNul=FALSE)
  line_num <- grep(pattern=pre_conc_text,current_file,ignore.case=FALSE,fixed=TRUE)
  
  # Makes sure only 1 line was identified.
  if(length(line_num)!=1){
    stop(paste("The user-defined beginning of the line of interest either does not match a line in file ",i," or it matches more than 1 line.",sep=""))
  }
  
  # Extracts max conc and rec. number from current line of interest.
  line_of_int <- strsplit(current_file[line_num],"\\s+")[[1]]
  current_max <- as.numeric(line_of_int[conc_pos])
  current_rec <- line_of_int[rec_pos]
  
  # Ensures units match user's expectations.
  if(expected_units!=line_of_int[units_pos]){
    stop(paste("Expected units of ",expected_units," do not match units in file ",i,".",sep=""))
  }
  
  # Builds data into output dataframe.
  out_data[i,1] <- current_lanes_or_skew
  out_data[i,2] <- current_grade
  out_data[i,3] <- current_rurality
  out_data[i,4] <- current_max
  out_data[i,5] <- current_rec
  
}

# Writes output data to file.
fwrite(out_data,file=output_path,sep=",",row.names=FALSE)

# This script was tested by taking a subset of the actual output data from the CAL3QHC NCHRP Task 104 
# modeling and manually checking that the results from the .ou2 files were correctly analyzed by this 
# script. Additional testing could be warrented.
#
# Hader; 