# This script minimally tested (spot-checking results, see end of script for details). Hader; 2018/09/21
#
# Slight modifications made on 1/2/2019 to account for incorporating varying gradients in IN files [Testing still needed].
#
# This script post-processes the .ou2 CAL3QHC modeling files
# for the interchange scenarios for the Task 104 project.
# The following analysis is performed on CAL3QHC output files:
#
# The user defines two .ou2 file directories: one for the 
# intersection scenario outputs, and one for the corresponding
# freeway scenario outputs.
# The output files from corresponding skew angles and urbanicity
# are mapped to eachother, and analyzed in tandem (this is done for all gradients
# available in the intersections files.
# Looping through the various lanes and distances of highway
# output files, the table of receptor- and wind direction-specific maximum
# concentrations are extracted from the intersection file 
# and the freeway file. For the freeway file, two sets of receptors
# are present in the outputs: one that corresponds to the freeway 
# being on the left side of the intersection, and one for it being
# on the right side.
# The intersection receptor concentration values are added to the 
# corresponding left-freeway and right-freeway receptors, in order
# to determine which receptor and wind direction corresponds to the
# highest overall concentration (from the sume of the freeway and intersection
# results).
# These maximum concentration results are then output to a .csv file that organizes values on
# the basis of left- or right-freeway, number of lanes in the freeway,
# distance between freeway and intersection, urban vs. rural classification, intersection gradient,
# and skew angle of the intersection.
#
# Note that this script has to be run separately for the different
# truck percentages (i.e., 10% and 20%).
#
# Author: John Hader

# Clears memory, console, and loads package(s).
rm(list = ls())
cat("\014")
require(data.table)

# User-defined inputs -----------------------------------------------------

# Input folder and output file path (remove .bat files from the CAL3QHC output modeling files folder).
#
# QA'ed: (JDH, 2019/01/15)
intersection_dir <- "C:/Users/37880/Desktop/NCHRP/Runs/10%_Trucks/Interchanges_2019/04_Output_files_IN" # No trailing backslash.
freeway_dir <- "C:/Users/37880/Desktop/NCHRP/Runs/10%_Trucks/Interchanges_2019/04_Output_files_FW" # No trailing backslash.
output_path <- "C:/Users/37880/Desktop/NCHRP/Runs/10%_Trucks/Interchanges_2019/Interchanges_10%_Results.csv" # As a .csv file.
file_name_sep <- "_" # Character that separates pieces of input file names.
file_ext <- ".ou2" # Include the '.' in the extension.

# Position of params in freeway and intersection file names (after strsplit using file_name_sep).
#
# Intersection
#
# QA'ed: (JDH, 2019/01/15)
skew_pos_int <- 2
grade_pos_int <- 3 # Assumed that "deg" preceeds the grade angle.
rurality_pos_int <- 4 # Assumed to be last param in file name and one char long (i.e., the ".ou2" is stripped out).
#
# Freeway
#
# QA'ed: (JDH, 2019/01/15)
skew_pos_free <- 2
grade_pos_free <- 3 # Assumed that "deg" preceeds the grade angle.
lane_pos_free <- 4 # Assumed that "ln" is at the end of this string.
dist_free <- 5 # Assumed that "ft" is at the end of this string.
rurality_pos_free <- 6 # Assumed to be last param in file name and one char long (i.e., the ".ou2" is stripped out).

# Defines expected grades for int. and freeway files.
# Note, include the leading "deg" and ending "%"
#
# QA'ed: (JDH, 2019/01/15)
expected_int_grade <- c("deg0%","deg1%","deg2%","deg3%","deg4%","deg5%","deg6%","deg7%")
expected_free_grade <- "deg0%" # Code currently only written to handle one of these gradients.

# Start/end lines of rec/wind dir. tables.
#
# QA'ed: (JDH, 2019/01/15)
freeway_lines <- c(65,100)
int_lines <- c(83,118)

# Columns to remove from rec/wind dir. tables (junk columns after string-splitting, namely the space column and the "*" column).
#
# QA'ed: (JDH, 2019/01/15)
junk_col <- c(1,3)

# Denotes column names from the output rec/wind dir. table.
#
# QA'ed: (JDH, 2019/01/15)
int_rec_wind_colnames <- c("Wind_ang","NE","NW","SE","SW")
free_rec_wind_colnames <-  c("Wind_ang","HW_Left_NE","HW_Left_NW","HW_Left_SE","HW_Left_SW",
                             "HW_Right_NE","HW_Right_NW","HW_Right_SE","HW_Right_SW")

# Columns for freeway receptors on the left and right sides of the intersection.
# Assumed that order of receptors through cols. matches that in intersections
# (i.e., defined in the two variables immediately above).
# Col. numbers here correspond to numbers after junk columns removed.
#
# QA'ed: (JDH, 2019/01/15)
free_col_left <- c(1:5)
free_col_right <- c(1,6:9)

#------ End User-defined Inputs--------

# Files pre-processing ----------------------------------------------------

# Reads in files from directory.
int_file_paths <- list.files(path=intersection_dir,full.names=TRUE)
int_file_names <- list.files(path=intersection_dir,full.names=FALSE)
#
free_file_paths <- list.files(path=freeway_dir,full.names=TRUE)
free_file_names <- list.files(path=freeway_dir,full.names=FALSE)

# Splits out character strings from input file names.
int_file_params <- matrix(unlist(strsplit(int_file_names,file_name_sep,fixed=TRUE)),ncol=length(strsplit(int_file_names,file_name_sep,fixed=TRUE)[[1]]),
                          nrow=length(strsplit(int_file_names,file_name_sep,fixed=TRUE)),byrow=TRUE)
free_file_params <- matrix(unlist(strsplit(free_file_names,file_name_sep,fixed=TRUE)),ncol=length(strsplit(free_file_names,file_name_sep,fixed=TRUE)[[1]]),
                           nrow=length(strsplit(free_file_names,file_name_sep,fixed=TRUE)),byrow=TRUE)

# Determines unique values of parameters present in int. and freeway files.
int_skews <- unique(as.numeric(int_file_params[,skew_pos_int]))
int_grades <- unique(int_file_params[,grade_pos_int])
int_URs <- unique(substr(int_file_params[,rurality_pos_int],1,1))
#
free_skews <- unique(as.numeric(free_file_params[,skew_pos_free]))
free_grades <- unique(free_file_params[,grade_pos_free])
free_lanes <- unique(free_file_params[,lane_pos_free])
free_dists <- unique(free_file_params[,dist_free])
free_URs <- unique(substr(free_file_params[,rurality_pos_free],1,1))

# Ensures that the parameters present match, where necessary, btwn int. and freeway files.
if((identical(int_skews[order(int_skews)],free_skews[order(free_skews)])==FALSE)|(identical(int_URs[order(int_URs)],free_URs[order(free_URs)])==FALSE)){
  stop("The intersection and freeway files must be composed of the same skew angles and urban/rural identifiers.")
}

# Ensures the gradients in the file names match expectations.
if((identical(expected_int_grade[order(expected_int_grade)],int_grades[order(int_grades)])==FALSE)|(identical(expected_free_grade,free_grades)==FALSE)){
  stop("The gradients in either the intersection and/or freeway file names don't match those anticipated by the user.")
}

# Files processing --------------------------------------------------------

# Creates dataframe for storing  conc. info.
num_combos <- 2*length(int_skews)*length(int_grades)*length(int_URs)*length(free_lanes)*length(free_dists) # '2' is for left vs. right FW.
out_data <- as.data.frame(matrix(data=NA,nrow=num_combos,ncol=9))
colnames(out_data) <- c("Skew_ang","Grade_Perc","Urban_or_Rural","Max_conc_PPM","Rec_Num","Wind_dir","Side_of_inter_freeway_is_on","Num_lanes","FW_int_distance_ft")

# Initializes the row counter.
row_counter <- 1

# Loops through the rurality identifiers.
for(i in c(1:length(int_URs))){
  
  # Defines current urban/rural parameter.
  current_UR <- int_URs[i]
 
  # Loops through the skew angles.
  for(j in c(1:length(int_skews))){
  
    # Defines current skew angle
    current_skew <- int_skews[j]  
    
    # Loops through the freeway # lanes.
    for(k in c(1:length(free_lanes))){
      
      # Defines current number of freeway lanes.
      current_lane <- free_lanes[k]
      
      # Loops through the # freeway/int. distances.
      for(l in c(1:length(free_dists))){
        
        # Defines current distance btwn freeway and intersection.
        current_dist <- free_dists[l]
        
        # Loops through the slope gradient angles (applies only to intersection files).
        for(m in c(1:length(int_grades))){
          
          # Defines the current slope gradient for the intersection file.
          current_grade <- int_grades[m]
          
          # Constructs the intersection and freeway file names for the current parameters.
          current_int_file <- paste(intersection_dir,"/",int_file_params[1,1],file_name_sep,current_skew,file_name_sep,current_grade,
                                    file_name_sep,current_UR,file_ext,sep="")
          current_free_file <- paste(freeway_dir,"/",free_file_params[1,1],file_name_sep,current_skew,file_name_sep,free_grades,
                                     file_name_sep,current_lane,file_name_sep,current_dist,file_name_sep,current_UR,file_ext,sep="")
          
          # Reads in the current files.
          current_int_data <- readLines(current_int_file,warn=TRUE,skipNul=FALSE)
          current_free_data <- readLines(current_free_file,warn=TRUE,skipNul=FALSE)
          
          # Extracts the receptor/wind dir. tables from each file as string character lines.
          int_table_lines <- unlist(current_int_data[int_lines[1]:int_lines[2]])
          free_table_lines <- unlist(current_free_data[freeway_lines[1]:freeway_lines[2]])
          
          # Converts string character lines of table to numeric (converted to matrix initially).
          int_table <- matrix(unlist(strsplit(int_table_lines,"\\s+",fixed=FALSE)),ncol=length(strsplit(int_table_lines,"\\s+",fixed=FALSE)[[1]]),
                              nrow=length(strsplit(int_table_lines,"\\s+",fixed=FALSE)),byrow=TRUE)
          free_table <- matrix(unlist(strsplit(free_table_lines,"\\s+",fixed=FALSE)),ncol=length(strsplit(free_table_lines,"\\s+",fixed=FALSE)[[1]]),
                               nrow=length(strsplit(free_table_lines,"\\s+",fixed=FALSE)),byrow=TRUE)
          
          # Formats the rec/wind dir. tables ('junk' columns removed and converted to data.table.
          int_table <- as.data.table(int_table[,-junk_col])
          free_table <- as.data.table(free_table[,-junk_col])
          colnames(int_table) <- int_rec_wind_colnames
          colnames(free_table) <- free_rec_wind_colnames 
          
          # Converts data.table elements to numeric (had been character up to this point)
          # enable conversion to matrix correctly.
          int_table <- as.matrix(int_table[,lapply(.SD,as.numeric)])
          free_table <- as.matrix(free_table[,lapply(.SD,as.numeric)])
          
          # Ensures there are no NA values left in the rec/wind dir. tables.
          if(any(is.na(int_table))|any(is.na(free_table))){
            stop("NA values were found within the rec/wind dir. tables.")
          }
          
          # Ensures the intersection and freeway tables contain the same wind dirs.
          if(identical(int_table[,1],free_table[,1])==FALSE){
            stop("The wind directions analyzed btwn the freeway and intersection must be identical.")
          }
          
          # Splits the freeway rec/wind dir. table into the left and right orientations.
          free_table_left <- free_table[,free_col_left]
          free_table_right <- free_table[,free_col_right]
          
          # Combines the intersection contributions to the left- and right-freeway layout receptors.
          total_left_free <- cbind(int_table[,1],int_table[,c(2:ncol(int_table))]+free_table_left[,c(2:ncol(free_table_left))])
          total_right_free <- cbind(int_table[,1],int_table[,c(2:ncol(int_table))]+free_table_right[,c(2:ncol(free_table_right))])
          #$$$$ First column doesn't have a name in above; is this an issue?
          
          # Extracts non-wind angle columns for use in determining maximum (and correponding receptor/wind angle).
          left_winds <- total_left_free[,1]
          left_data <- total_left_free[,c(2:ncol(total_left_free))]
          right_winds <- total_right_free[,1]
          right_data <- total_right_free[,c(2:ncol(total_right_free))]
          
          # Determines the maximum conc. and corresponding receptor/wind direction.
          left_free_max <- max(left_data)
          left_free_wind <- left_winds[which(left_data==left_free_max,arr.in=TRUE)[1,1]]
          left_free_rec <- colnames(left_data)[which(left_data==left_free_max,arr.in=TRUE)[1,2]]
          #
          right_free_max <- max(right_data)
          right_free_wind <- right_winds[which(right_data==right_free_max,arr.in=TRUE)[1,1]]
          right_free_rec <- colnames(right_data)[which(right_data==right_free_max,arr.in=TRUE)[1,2]]

          # Builds current parameter information into output table.
          #
          # Left freeway data.
          out_data[row_counter,"Skew_ang"] <- current_skew
          out_data[row_counter,"Grade_Perc"] <- paste("Freeway ",substr(free_grades,4,nchar(free_grades))," Intersection ",substr(current_grade,4,nchar(current_grade)),sep="")
          out_data[row_counter,"Urban_or_Rural"] <- current_UR
          out_data[row_counter,"Max_conc_PPM"] <- left_free_max
          out_data[row_counter,"Rec_Num"] <- left_free_rec
          out_data[row_counter,"Wind_dir"] <- left_free_wind
          out_data[row_counter,"Side_of_inter_freeway_is_on"] <- "Left"
          out_data[row_counter,"Num_lanes"] <- current_lane
          out_data[row_counter,"FW_int_distance_ft"] <- current_dist
          
          #
          row_counter <- row_counter+1
          
          # Right freeway data.
          out_data[row_counter,"Skew_ang"] <- current_skew
          out_data[row_counter,"Grade_Perc"] <- paste("Freeway ",substr(free_grades,4,nchar(free_grades))," Intersection ",substr(current_grade,4,nchar(current_grade)),sep="")
          out_data[row_counter,"Urban_or_Rural"] <- current_UR
          out_data[row_counter,"Max_conc_PPM"] <- right_free_max
          out_data[row_counter,"Rec_Num"] <- right_free_rec
          out_data[row_counter,"Wind_dir"] <- right_free_wind
          out_data[row_counter,"Side_of_inter_freeway_is_on"] <- "Right"
          out_data[row_counter,"Num_lanes"] <- current_lane
          out_data[row_counter,"FW_int_distance_ft"] <- current_dist
          
          # Informs the user what row of output data is currently being processed.
          if((row_counter%%50)==0){
            print(paste("Finished processing row ",row_counter,"/",num_combos,sep=""))
          }
          
          # Increments row counter.
          row_counter <- row_counter+1
          
        }  
        
      }
      
    }
    
  }

}

# Writes output data to file.
fwrite(out_data,file=output_path,sep=",",row.names=FALSE)

# This script was re-read after initial writing by Hader. The script was run on the 10% Trucks
# data, and 3 scenarios were manually analyzed to produce the maximum concentration and
# corresponding receptor/wind direction for both the HW Right and HW Left scenarios.
# Manually-generated results from all 3 of these scenarios matched the script-generated 
# values.
#
# Hader;